@@ -1,4 +1,5 @@
 from __future__ import absolute_import
+from __future__ import division
 _last_figure = None
 import ipywidgets
 from IPython.display import display
@@ -13,6 +14,8 @@ from . import examples
 import warnings
 import PIL.Image
 import traitlets
+import uuid
+
 try:
     from io import BytesIO as StringIO
 except:
@@ -30,12 +33,12 @@ _seq_snm = "If an (S, N, M) array, the first dimension will be used for frames i
 
 _doc_snippets = {}
 _doc_snippets[
-    "color"] = "color for each point/vertex/symbol, can be string format, examples for red:'red', '#f00', '#ff0000' or 'rgb(1,0,0), or rgb array of shape (N, 3) or (S, N, 3)"
+    "color"] = "color for each point/vertex/symbol, can be string format, examples for red:'red', '#f00', '#ff0000' or 'rgb(1,0,0), or rgb array of shape (N, 3 or 4) or (S, N, 3 or 4)"
 _doc_snippets[
-    "color2d"] = "color for each point/vertex string format, examples for red:'red', '#f00', '#ff0000' or 'rgb(1,0,0), or rgb array of shape (2, N, 3) or (S, 2, N, 3)"
+    "color2d"] = "color for each point/vertex string format, examples for red:'red', '#f00', '#ff0000' or 'rgb(1,0,0), or rgb array of shape (2, N, 3 or 4) or (S, 2, N, 3 or 4)"
 _doc_snippets[
     "size"] = "float representing the size of the glyph in percentage of the viewport, where 100 is the full size of the viewport"
-_doc_snippets["marker"] = "name of the marker, options are: 'arrow', 'box', 'diamond', 'sphere'"
+_doc_snippets["marker"] = "name of the marker, options are: 'arrow', 'box', 'diamond', 'sphere', 'point_2d', 'square_2d', 'triangle_2d', 'circle_2d'"
 _doc_snippets["x"] = "numpy array of shape (N,) or (S, N) with x positions. {}".format(_seq_sn)
 _doc_snippets["y"] = "idem for y"
 _doc_snippets["z"] = "idem for z"
@@ -61,8 +64,26 @@ def clear():
     current.container = None
     current.figure = None
 
+def controls_light(return_widget=False):
+    fig = gcf()
+    ambient_coefficient = ipywidgets.FloatSlider(min=0, max=1, step=0.001, value=fig.ambient_coefficient,
+                                                 description="ambient")
+    diffuse_coefficient = ipywidgets.FloatSlider(min=0, max=1, step=0.001, value=fig.diffuse_coefficient,
+                                                 description="diffuse")
+    specular_coefficient = ipywidgets.FloatSlider(min=0, max=1, step=0.001, value=fig.specular_coefficient,
+                                                  description="specular")
+    specular_exponent = ipywidgets.FloatSlider(min=0, max=10, step=0.001, value=fig.specular_exponent,
+                                               description="specular exp")
+    ipywidgets.jslink((fig, 'ambient_coefficient'), (ambient_coefficient, 'value'))
+    ipywidgets.jslink((fig, 'diffuse_coefficient'), (diffuse_coefficient, 'value'))
+    ipywidgets.jslink((fig, 'specular_coefficient'), (specular_coefficient, 'value'))
+    ipywidgets.jslink((fig, 'specular_exponent'), (specular_exponent, 'value'))
+    widgets_bottom = [ipywidgets.HBox([ambient_coefficient, diffuse_coefficient]),
+                      ipywidgets.HBox([specular_coefficient, specular_exponent])]
+    current.container.children += tuple(widgets_bottom, )
+    if return_widget: return widgets_bottom
 
-def figure(key=None, width=400, height=500, lighting=True, controls=True, controls_vr=False, debug=False, **kwargs):
+def figure(key=None, width=400, height=500, lighting=True, controls=True, controls_vr=False, controls_light=False, debug=False, **kwargs):
     """Create a new figure (if no key is given) or return the figure associated with key
 
     :param key: Python object that identifies this figure
@@ -77,13 +98,19 @@ def figure(key=None, width=400, height=500, lighting=True, controls=True, contro
     if key is not None and key in current.figures:
         current.figure = current.figures[key]
         current.container = current.containers[key]
+    elif isinstance(key, ipv.Figure) and key in current.figures.values():
+        key_index = list(current.figures.values()).index(key)
+        key = list(current.figures.keys())[key_index]
+        current.figure = current.figures[key]
+        current.container = current.containers[key]
     else:
-        current.figure = ipv.Figure(volume_data=None, width=width, height=height, **kwargs)
+        current.figure = ipv.Figure(width=width, height=height, **kwargs)
         current.container = ipywidgets.VBox()
         current.container.children = [current.figure]
-        if key is not None:
-            current.figures[key] = current.figure
-            current.containers[key] = current.container
+        if key is None:
+            key = uuid.uuid4().hex
+        current.figures[key] = current.figure
+        current.containers[key] = current.container
         if controls:
             #stereo = ipywidgets.ToggleButton(value=current.figure.stereo, description='stereo', icon='eye')
             #l1 = ipywidgets.jslink((current.figure, 'stereo'), (stereo, 'value'))
@@ -93,8 +120,10 @@ def figure(key=None, width=400, height=500, lighting=True, controls=True, contro
             eye_separation = ipywidgets.FloatSlider(value=current.figure.eye_separation, min=-10, max=10, icon='eye')
             ipywidgets.jslink((eye_separation, 'value'), (current.figure, 'eye_separation'))
             current.container.children += (eye_separation,)
+        if controls_light:
+            globals()['controls_light']()
         if debug:
-            show = ipywidgets.ToggleButtons(options=["Volume", "Back", "Front"])
+            show = ipywidgets.ToggleButtons(options=["Volume", "Back", "Front", "Coordinate"])
             current.container.children += (show,)
             #ipywidgets.jslink((current.figure, 'show'), (show, 'value'))
             traitlets.link((current.figure, 'show'), (show, 'value'))
@@ -113,9 +142,18 @@ def gcf():
 
 
 def _grow_limit(limits, values):
-    finites = np.isfinite(values)
-    newvmin = np.min(values[finites])
-    newvmax = np.max(values[finites])
+    if isinstance(values, (tuple, list)) and len(values) == 2:
+        newvmin, newvmax = values
+    else:
+        try:
+            values[0]  # test if scalar
+        except TypeError:
+            newvmin = values
+            newvmax = values
+        else:
+            finites = np.isfinite(values)
+            newvmin = np.min(values[finites])
+            newvmax = np.max(values[finites])
     if limits is None:
         return newvmin, newvmax
     else:
@@ -295,62 +333,40 @@ def plot_mesh(x, y, z, color=default_color, wireframe=True, surface=True, wrapx=
     else:
         nx, ny = shape = x[0].shape
 
+    # assert len(x.shape) == 2, "Array x must be 2 dimensional."
+    # assert len(y.shape) == 2, "Array y must be 2 dimensional."
+    # assert len(z.shape) == 2, "Array z must be 2 dimensional."
+    # assert x.shape == y.shape, "Arrays x and y must have same shape."
+    # assert y.shape == z.shape, "Arrays y and z must have same shape."
+    # convert x, y, z from shape (nx, ny) to (nx * ny) or
+    # (frame, nx, ny) to (frame, nx*ny)
     def reshape(ar):
         if dim(ar) == 3:
             return [k.reshape(-1) for k in ar]
         else:
             return ar.reshape(-1)
-
-    def reshape_color(ar):
-        if dim(ar) == 4:
-            return [k.reshape(-1, 3) for k in ar]
-        else:
-            return ar.reshape(-1, 3)
-
-    if isinstance(color, np.ndarray):
-        # if dim(color) == 4:
-        #	color = color.reshape((color.shape[0], -1, color.shape[-1]))
-        color = reshape_color(color)
-        # print(color.shape)
-
     x = reshape(x)
     y = reshape(y)
     z = reshape(z)
+    # similar for texture coordinates
     if u is not None:
         u = reshape(u)
     if v is not None:
         v = reshape(v)
+
+    # convert color from shape (nx, ny, {3,4}) to (nx * ny, {3, 4}) or
+    # (frame, nx, ny, {3,4}) to (frame, nx*ny, {3,4})
+    def reshape_color(ar):
+        if dim(ar) == 4:
+            return [k.reshape(-1, k.shape[-1]) for k in ar]
+        else:
+            return ar.reshape(-1, ar.shape[-1])
+
+    if isinstance(color, np.ndarray):
+        color = reshape_color(color)
+
     _grow_limits(np.array(x).reshape(-1), np.array(y).reshape(-1), np.array(z).reshape(-1))
-    mx = nx if wrapx else nx - 1
-    my = ny if wrapy else ny - 1
-    triangles = np.zeros(((mx) * (my) * 2, 3), dtype=np.uint32)
-    lines = np.zeros(((mx) * (my) * 4, 2), dtype=np.uint32)
-
-    def index_from2d(i, j):
-        xi = (i % nx)
-        yi = (j % ny)
-        return ny * xi + yi
-        """
-        ^ ydir
-        |
-        2 3
-        0 1  ---> x dir
-        """
-
-    for i in range(mx):
-        for j in range(my):
-            p0 = index_from2d(i, j)
-            p1 = index_from2d(i + 1, j)
-            p2 = index_from2d(i, j + 1)
-            p3 = index_from2d(i + 1, j + 1)
-            triangle_index = (i * my) + j
-            triangles[triangle_index * 2 + 0, :] = [p0, p1, p3]
-            triangles[triangle_index * 2 + 1, :] = [p0, p3, p2]
-            lines[triangle_index * 4 + 0, :] = [p0, p1]
-            lines[triangle_index * 4 + 1, :] = [p0, p2]
-            lines[triangle_index * 4 + 2, :] = [p2, p3]
-            lines[triangle_index * 4 + 3, :] = [p1, p3]
-    # print(i, j, p0, p1, p2, p3)
+    triangles, lines = _make_triangles_lines((nx,ny) ,wrapx,wrapy)
     mesh = ipv.Mesh(x=x, y=y, z=z, triangles=triangles if surface else None, color=color,
                        lines=lines if wireframe else None,
                        u=u, v=v, texture=texture)
@@ -374,6 +390,7 @@ def plot(x, y, z, color=default_color, **kwargs):
                     size=1, connected=True, visible_markers=False)
     kwargs = dict(defaults, **kwargs)
     s = ipv.Scatter(x=x, y=y, z=z, color=color, **kwargs)
+    s.material.visible = False
     fig.scatters = fig.scatters + [s]
     return s
 
@@ -382,7 +399,7 @@ def plot(x, y, z, color=default_color, **kwargs):
 def scatter(x, y, z, color=default_color, size=default_size,
             size_selected=default_size_selected,
             color_selected=default_color_selected, marker="diamond",
-            selection=None, **kwargs):
+            selection=None, grow_limits=True, **kwargs):
     """Plots many markers/symbols in 3d
 
     :param x: {x}
@@ -398,7 +415,8 @@ def scatter(x, y, z, color=default_color, size=default_size,
     :return: :any:`Scatter`
     """
     fig = gcf()
-    _grow_limits(x, y, z)
+    if grow_limits:
+        _grow_limits(x, y, z)
     s = ipv.Scatter(x=x, y=y, z=z, color=color, size=size,
                     color_selected=color_selected, size_selected=size_selected,
                     geo=marker, selection=selection, **kwargs)
@@ -471,12 +489,29 @@ def animation_control(object, sequence_length=None, add=True, interval=200):
         # get all non-None arrays
         sequence_lengths = []
         for object in objects:
+            sequence_lengths_previous = list(sequence_lengths)
             values = [getattr(object, name) for name in "x y z vx vy vz".split() if hasattr(object, name)]
             values = [k for k in values if k is not None]
             # sort them such that the higest dim is first
             values.sort(key=lambda key: -len(key.shape))
-            sequence_length = values[0].shape[0]  # assume this defines the sequence length
-            sequence_lengths.append(sequence_length)
+            try:
+                sequence_length = values[0].shape[0]  # assume this defines the sequence length
+                if isinstance(object, ipv.Mesh):  # for a mesh, it does not make sense to have less than 1 dimension
+                    if len(values[0].shape) >= 2:  # if just 1d, it is most likely not an animation
+                        sequence_lengths.append(sequence_length)
+                else:
+                    sequence_lengths.append(sequence_length)
+            except IndexError:      # scalars get ignored
+                pass
+            if hasattr(object, 'color'):
+                color = object.color
+                if color is not None:
+                    shape = color.shape
+                    if len(shape) == 3:  # would be the case for for (frame, point_index, color_index)
+                        sequence_lengths.append(shape[0])
+                    # TODO: maybe support arrays of string type of form (frame, point_index)
+            if len(sequence_lengths) == len(sequence_lengths_previous):
+                raise ValueError('no frame dimension found for object: {}'.format(object))
         sequence_length = max(sequence_lengths)
     fig = gcf()
     fig.animation = interval
@@ -537,15 +572,18 @@ def transfer_function(level=[0.1, 0.5, 0.9], opacity=[0.01, 0.05, 0.1], level_wi
         current.container.children = (tf.control(max_opacity=max_opacity),) + current.container.children
     return tf
 
-def plot_isosurface(data, level=None, color=default_color, wireframe=True, surface=True, controls=True):
+
+def plot_isosurface(data, level=None, color=default_color, wireframe=True, surface=True, controls=True, extent=None):
     """Plot a surface at constant value (like a 2d contour)
 
+    :param data: 3d numpy array
     :param float level: value where the surface should lie
     :param color: color of the surface, although it can be an array, the length is difficult to predict beforehand,
                   if per vertex color are needed, it is better to set them on the returned mesh afterwards.
     :param bool wireframe: draw lines between the vertices
     :param bool surface: draw faces/triangles between the vertices
     :param bool controls: add controls to change the isosurface
+    :param extent: list of [[xmin, xmax], [ymin, ymax], [zmin, zmax]] values that define the bounding box of the mesh, otherwise the viewport is used
     :return: :any:`Mesh`
     """
     from skimage import measure
@@ -559,6 +597,15 @@ def plot_isosurface(data, level=None, color=default_color, wireframe=True, surfa
     # in the future we may want to support normals and the values (with colormap)
     # and require skimage >= 0.13
     x, y, z = verts.T
+
+    # Rescale coordinates to given limits
+    if extent:
+        xlim, ylim, zlim = extent
+        x = x * np.diff(xlim)/(data.shape[0]-1) + xlim[0]
+        y = y * np.diff(ylim)/(data.shape[1]-1) + ylim[0]
+        z = z * np.diff(zlim)/(data.shape[2]-1) + zlim[0]
+        _grow_limits(*extent)
+
     mesh = plot_trisurf(x, y, z, triangles=triangles, color=color)
     if controls:
         vmin, vmax = np.percentile(data, 1),  np.percentile(data, 99)
@@ -589,21 +636,25 @@ def plot_isosurface(data, level=None, color=default_color, wireframe=True, surfa
     return mesh
 
 
-def volshow(data, lighting=False, data_min=None, data_max=None, tf=None, stereo=False,
+def volshow(data, lighting=False, data_min=None, data_max=None,
+            max_shape=256, tf=None, stereo=False,
             ambient_coefficient=0.5, diffuse_coefficient=0.8,
             specular_coefficient=0.5, specular_exponent=5,
             downscale=1,
             level=[0.1, 0.5, 0.9], opacity=[0.01, 0.05, 0.1], level_width=0.1,
-            controls=True, max_opacity=0.2):
+            controls=True, max_opacity=0.2, memorder='C', extent=None):
     """Visualize a 3d array using volume rendering.
 
     Currently only 1 volume can be rendered.
 
 
     :param data: 3d numpy array
+    :param origin: origin of the volume data, this is to match meshes which have a different origin
+    :param domain_size: domain size is the size of the volume 
     :param bool lighting: use lighting or not, if set to false, lighting parameters will be overriden
     :param float data_min: minimum value to consider for data, if None, computed using np.nanmin
     :param float data_max: maximum value to consider for data, if None, computed using np.nanmax
+    :parap int max_shape: maximum shape for the 3d cube, if larger, the data is reduced by skipping/slicing (data[::N]), set to None to disable.
     :param tf: transfer function (or a default one)
     :param bool stereo: stereo view for virtual reality (cardboard and similar VR head mount)
     :param ambient_coefficient: lighting parameter
@@ -616,43 +667,54 @@ def volshow(data, lighting=False, data_min=None, data_max=None, tf=None, stereo=
     :param level_width: width of the (gaussian) bumps where the opacity peaks, scalar or sequence of max length 3
     :param bool controls: add controls for lighting and transfer function or not
     :param float max_opacity: maximum opacity for transfer function controls
+    :param extent: list of [[xmin, xmax], [ymin, ymax], [zmin, zmax]] values that define the bounds of the volume, otherwise the viewport is used
     :return:
     """
-    vol = gcf()
+    fig = gcf()
+
     if tf is None:
-        tf = vol.tf or transfer_function(level, opacity, level_width, controls=controls, max_opacity=max_opacity)
+        tf = transfer_function(level, opacity, level_width, controls=controls, max_opacity=max_opacity)
     if data_min is None:
         data_min = np.nanmin(data)
     if data_max is None:
         data_max = np.nanmax(data)
-    vol.tf = tf
-    vol.data_min = data_min
-    vol.data_max = data_max
-    vol.volume_data = data
-    vol.stereo = stereo
-    vol.ambient_coefficient = ambient_coefficient
-    vol.diffuse_coefficient = diffuse_coefficient
-    vol.specular_coefficient = specular_coefficient
-    vol.specular_exponent = specular_exponent
+    if memorder is 'F':
+        data = data.T
+
+    if extent is None:
+        extent = [(0, k) for k in data.shape[::-1]]
+
+    if extent:
+        _grow_limits(*extent)
+
+    vol = ipv.Volume(data_original = data,
+                    tf=tf,
+                    data_min = data_min,
+                    data_max = data_max,
+                    show_min = data_min,
+                    show_max = data_max,
+                    extent_original = extent,
+                    data_max_shape = max_shape,
+                    ambient_coefficient = ambient_coefficient,
+                    diffuse_coefficient = diffuse_coefficient,
+                    specular_coefficient = specular_coefficient,
+                    specular_exponent = specular_exponent,
+                    rendering_lighting = lighting)
+
+    vol._listen_to(fig)
 
     if controls:
-        ambient_coefficient = ipywidgets.FloatSlider(min=0, max=1, step=0.001, value=vol.ambient_coefficient,
-                                                     description="ambient")
-        diffuse_coefficient = ipywidgets.FloatSlider(min=0, max=1, step=0.001, value=vol.diffuse_coefficient,
-                                                     description="diffuse")
-        specular_coefficient = ipywidgets.FloatSlider(min=0, max=1, step=0.001, value=vol.specular_coefficient,
-                                                      description="specular")
-        specular_exponent = ipywidgets.FloatSlider(min=0, max=10, step=0.001, value=vol.specular_exponent,
-                                                   description="specular exp")
-        # angle2 = ipywidgets.FloatSlider(min=0, max=np.pi*2, value=v.angle2, description="angle2")
-        ipywidgets.jslink((vol, 'ambient_coefficient'), (ambient_coefficient, 'value'))
-        ipywidgets.jslink((vol, 'diffuse_coefficient'), (diffuse_coefficient, 'value'))
-        ipywidgets.jslink((vol, 'specular_coefficient'), (specular_coefficient, 'value'))
-        ipywidgets.jslink((vol, 'specular_exponent'), (specular_exponent, 'value'))
-        widgets_bottom = [ipywidgets.HBox([ambient_coefficient, diffuse_coefficient]),
-                          ipywidgets.HBox([specular_coefficient, specular_exponent])]
+        widget_opacity_scale = ipywidgets.FloatLogSlider(base=10, min=-2, max=2, 
+                                                     description="opacity")
+        widget_brightness = ipywidgets.FloatLogSlider(base=10, min=-1, max=1,
+                                                     description="brightness")
+        ipywidgets.jslink((vol, 'opacity_scale'), (widget_opacity_scale, 'value'))
+        ipywidgets.jslink((vol, 'brightness'), (widget_brightness, 'value'))
+        widgets_bottom = [ipywidgets.HBox([widget_opacity_scale, widget_brightness])]
         current.container.children += tuple(widgets_bottom, )
 
+    fig.volumes = fig.volumes + [vol]
+
     return vol
 
 
@@ -684,11 +746,12 @@ def save(filepath, makedirs=True, title=u'IPyVolume Widget', all_states=False,
                                offline_cors=offline_cors)
 
 
-def _change_y_angle(fig, frame, fraction):
-    fig.angley = fraction * np.pi * 2
+def _change_azimuth_angle(fig, frame, fraction):
+    with fig:
+        view(azimuth=fraction*360)
 
 
-def movie(f="movie.mp4", function=_change_y_angle, fps=30, frames=30, endpoint=False, \
+def movie(f="movie.mp4", function=_change_azimuth_angle, fps=30, frames=30, endpoint=False, \
           cmd_template_ffmpeg="ffmpeg -y -r {fps} -i {tempdir}/frame-%5d.png -vcodec h264 -pix_fmt yuv420p {filename}",
           cmd_template_gif="convert -delay {delay} {loop} {tempdir}/frame-*.png {filename}",
           gif_loop=0):
@@ -744,7 +807,7 @@ def movie(f="movie.mp4", function=_change_y_angle, fps=30, frames=30, endpoint=F
     return tempdir
 
 
-def _screenshot_data(timeout_seconds=10, output_widget=None, format="png", width=None, height=None, fig=None, headless=False):
+def _screenshot_data(timeout_seconds=10, output_widget=None, format="png", width=None, height=None, fig=None, headless=False, devmode=False):
     if fig is None:
         fig = gcf()
     else:
@@ -754,8 +817,10 @@ def _screenshot_data(timeout_seconds=10, output_widget=None, format="png", width
         import tempfile
         tempdir = tempfile.mkdtemp()
         tempfile = os.path.join(tempdir, 'headless.html')
-        save(tempfile, offline=True, scripts_path=tempdir, devmode=True)
+        save(tempfile, offline=False, scripts_path=tempdir, devmode=devmode)
         data = headless._screenshot_data("file://" + tempfile)
+        if data is None:
+            raise ValueError('Error capturing data from headless browser')
     else:
         if output_widget is None:
             output_widget = ipywidgets.Output()
@@ -792,7 +857,7 @@ def _screenshot_data(timeout_seconds=10, output_widget=None, format="png", width
     data = data[data.find(",") + 1:]
     return base64.b64decode(data)
 
-def screenshot(width=None, height=None, format="png", fig=None, timeout_seconds=10, output_widget=None, headless=False):
+def screenshot(width=None, height=None, format="png", fig=None, timeout_seconds=10, output_widget=None, headless=False, devmode=False):
     """Save the figure to a PIL.Image object.
 
     :param int width: the width of the image in pixels
@@ -804,16 +869,18 @@ def screenshot(width=None, height=None, format="png", fig=None, timeout_seconds=
     :param timeout_seconds: maximum time to wait for image data to return
     :type output_widget: ipywidgets.Output
     :param output_widget: a widget to use as a context manager for capturing the data
+    :param bool headless: if True, use headless chrome to take screenshot
+    :param bool devmode: if True, attempt to get index.js from local js/dist folder
     :return: PIL.Image
 
     """
     assert format in ['png','jpeg','svg'], "image format must be png, jpeg or svg"
     data = _screenshot_data(timeout_seconds=timeout_seconds, output_widget=output_widget,
-    format=format, width=width, height=height, fig=fig, headless=headless)
+    format=format, width=width, height=height, fig=fig, headless=headless, devmode=devmode)
     f = StringIO(data)
     return PIL.Image.open(f)
 
-def savefig(filename, width=None, height=None, fig=None, timeout_seconds=10, output_widget=None, headless=False):
+def savefig(filename, width=None, height=None, fig=None, timeout_seconds=10, output_widget=None, headless=False, devmode=False):
     """Save the figure to an image file.
 
     :param str filename: must have extension .png, .jpeg or .svg
@@ -823,13 +890,15 @@ def savefig(filename, width=None, height=None, fig=None, timeout_seconds=10, out
     :param fig: if None use the current figure
     :param float timeout_seconds: maximum time to wait for image data to return
     :param ipywidgets.Output output_widget: a widget to use as a context manager for capturing the data
+    :param bool headless: if True, use headless chrome to save figure
+    :param bool devmode: if True, attempt to get index.js from local js/dist folder
     """
     __, ext = os.path.splitext(filename)
     format = ext[1:]
     assert format in ['png','jpeg','svg'], "image format must be png, jpeg or svg"
     with open(filename, "wb") as f:
         f.write(_screenshot_data(timeout_seconds=timeout_seconds, output_widget=output_widget,
-        format=format, width=width, height=height, fig=fig, headless=headless))
+        format=format, width=width, height=height, fig=fig, headless=headless, devmode=devmode))
 
 
 def xlabel(label):
@@ -856,16 +925,31 @@ def xyzlabel(labelx, labely, labelz):
     ylabel(labely)
     zlabel(labelz)
 
-def view(azimuth, elevation):
-    """Sets camera angles
+def view(azimuth=None, elevation=None, distance=None):
+    """Sets camera angles and distance and returns the current.
 
     :param float azimuth: rotation around the axis pointing up in degrees
     :param float elevation: rotation where +90 means 'up', -90 means 'down', in degrees
+    :param float distance: radial distance from the center to the camera.
     """
     fig = gcf()
-    fig.anglex = np.radians(elevation)
-    fig.angley = np.radians(azimuth)
-
+    # first calculate the current values
+    x, y, z = fig.camera.position
+    r = np.sqrt(x**2 + y**2 + z**2)
+    az = np.degrees(np.arctan2(x, z))
+    el = np.degrees(np.arcsin(y/r))
+    if azimuth is None:
+        azimuth = az
+    if elevation is None:
+        elevation = el
+    if distance is None:
+        distance = r
+    cosaz = np.cos(np.radians(azimuth))
+    sinaz = np.sin(np.radians(azimuth))
+    sine = np.sin(np.radians(elevation))
+    cose = np.cos(np.radians(elevation))
+    fig.camera.position = (distance*sinaz*cose, distance*sine, distance*cosaz*cose)
+    return azimuth, elevation, distance
 
 
 # mimic matplotlib namesace
@@ -873,10 +957,11 @@ class style:
     """'Static class that mimics a matplotlib module.
 
     Example:
-    >>> import ipyvolume.pylab as p3
-    >>> p3.style.use('light'])
-    >>> p3.style.use('seaborn-darkgrid'])
-    >>> p3.style.use(['seaborn-darkgrid', {'axes.x.color':'orange'}])
+
+    >>> import ipyvolume as ipv
+    >>> ipv.style.use('light'])
+    >>> ipv.style.use('seaborn-darkgrid'])
+    >>> ipv.style.use(['seaborn-darkgrid', {'axes.x.color':'orange'}])
 
     Possible style values:
      * figure.facecolor: background color
@@ -922,11 +1007,12 @@ class style:
                     utils.nested_setitem(style, to_name, value)
             return style
 
-        if isinstance(style, six.string_types):
+        if isinstance(style, six.string_types + (dict,)):
             styles = [style]
         else:
             styles = style
-        totalstyle = utils.dict_deep_update({}, ipyvolume.styles._defaults)
+        fig = gcf()
+        totalstyle = utils.dict_deep_update({}, fig.style)
         for style in styles:
             if isinstance(style, six.string_types):
                 if hasattr(ipyvolume.styles, style):
@@ -947,18 +1033,54 @@ class style:
         fig = gcf()
         fig.style = totalstyle
 
+    @staticmethod
+    def axes_off():
+        """Do not draw the axes"""
+        style.use({'axes': {'visible': False}})
+
+    @staticmethod
+    def axes_on():
+        """Draw the axes"""
+        style.use({'axes': {'visible': True}})
+
+    @staticmethod
+    def box_off():
+        """Do not draw the box around the visible volume"""
+        style.use({'box': {'visible': False}})
+
+    @staticmethod
+    def box_on():
+        """Draw a box around the visible volume"""
+        style.use({'box': {'visible': True}})
+
+    @staticmethod
+    def background_color(color):
+        """Sets the background color"""
+        style.use({'background-color': color})
+
+for style_name, __ in ipv.styles.styles.items():
+    def closure(style_name=style_name):
+        def quick_set():
+            style.use(style_name)
+        attr_name = 'set_style_' + style_name
+        attr = staticmethod(quick_set)
+        setattr(style, attr_name, attr)
+        getattr(style, attr_name).__doc__ = """Short for style.use(%r)""" % style_name
+
+    closure()
+
 @_docsubst
 def plot_plane(where="back", texture=None):
     """Plots a plane at a particular location in the viewbox
 
-    :param str where: 'back', 'front', 'left', 'right'
+    :param str where: 'back', 'front', 'left', 'right', 'top', 'bottom'
     :param texture: {texture}
     :return: :any:`Mesh`
     """
     fig = gcf()
     xmin, xmax = fig.xlim
     ymin, ymax = fig.ylim
-    zmin, zmax = fig.ylim
+    zmin, zmax = fig.zlim
     if where == "back":
         x = [xmin, xmax, xmax, xmin]
         y = [ymin, ymin, ymax, ymax]
@@ -968,13 +1090,21 @@ def plot_plane(where="back", texture=None):
         y = [ymin, ymin, ymax, ymax]
         z = [zmax, zmax, zmax, zmax]
     if where == "left":
-        x = [xmin, xmin, xmin, zmin]
+        x = [xmin, xmin, xmin, xmin]
         y = [ymin, ymin, ymax, ymax]
         z = [zmin, zmax, zmax, zmin]
     if where == "right":
-        x = [xmax, xmax, xmax, zmax]
+        x = [xmax, xmax, xmax, xmax]
         y = [ymin, ymin, ymax, ymax]
         z = [zmin, zmax, zmax, zmin][::-1]
+    if where == "top":
+        x = [xmin, xmax, xmax, xmin]
+        y = [ymax, ymax, ymax, ymax]
+        z = [zmax, zmax, zmin, zmin]
+    if where == "bottom":
+        x = [xmax, xmin, xmin, xmax]
+        y = [ymin, ymin, ymin, ymin]
+        z = [zmin, zmin, zmax, zmax]
     triangles = [(0, 1, 2), (0, 2, 3)]
     u = v = None
     if texture is not None:
@@ -983,24 +1113,157 @@ def plot_plane(where="back", texture=None):
     mesh = plot_trisurf(x, y, z, triangles, texture=texture, u=u, v=v)
     return mesh
 
-def selector_lasso(output_widget=None):
+def selector_default(output_widget=None):
+    """Capture selection events from the current figure, and apply the selections to Scatter objects
+
+    Example:
+
+    >>> import ipyvolume as ipv
+    >>> ipv.figure()
+    >>> ipv.examples.gaussian()
+    >>> ipv.selector_default()
+    >>> ipv.show()
+
+    Now hold the control key to do selections, type
+
+      * 'C' for circle
+      * 'R' for rectangle
+      * 'L' for lasso
+      * '=' for replace mode
+      * '&' for logically and mode
+      * '|' for logically or mode
+      * '-' for subtract mode
+
+    """
     fig = gcf()
     if output_widget is None:
         output_widget = ipywidgets.Output()
         display(output_widget)
     def lasso(data, other=None, fig=fig):
         with output_widget:
-            if data['device']:
+            if data['device'] and data['type'] == 'lasso':
                 import shapely.geometry
                 region = shapely.geometry.Polygon(data['device'])
-                for scatter in fig.scatters:
-                    xyz_projected = fig.project(scatter.x, scatter.y, scatter.z)
-                    points = xyz_projected.T[:,:2]
-                    selected = []
-                    for i, p in enumerate(points):
-                        #print(i, p)
-                        if region.contains(shapely.geometry.Point(p)):
-                            selected.append(i)
-                    if selected:
-                        scatter.selected = selected
-    fig.on_lasso(lasso)
+                @np.vectorize
+                def inside(x, y):
+                    return region.contains(shapely.geometry.Point([x, y]))
+            if data['device'] and data['type'] == 'circle':
+                x1, y1 = data['device']['begin']
+                x2, y2 = data['device']['end']
+                dx = x2 - x1
+                dy = y2 - y1
+                r = (dx**2 + dy**2)**0.5
+                def inside(x, y):
+                    return ((x-x1)**2 + (y-y1)**2) < r**2
+            if data['device'] and data['type'] == 'rectangle':
+                x1, y1 = data['device']['begin']
+                x2, y2 = data['device']['end']
+                x = [x1, x2]
+                y = [y1, y2]
+                xmin, xmax = min(x), max(x)
+                ymin, ymax = min(y), max(y)
+                def inside(x, y):
+                    return (x > xmin) & (x < xmax) & (y > ymin) & (y < ymax)
+            def join(x, y, mode):
+                N = np.max(x) if x is not None else np.max(y)
+                N = max(N, np.max(y))
+                xmask = np.zeros(N+1, np.bool)
+                ymask = np.zeros(N+1, np.bool)
+                if x is not None:
+                    xmask[x] = True
+                ymask[y] = True
+                if mode == "replace":
+                    return np.where(ymask)
+                if mode == "and":
+                    mask = xmask & ymask
+                    return np.where(ymask if x is None else mask)
+                if mode == "or":
+                    mask = xmask | ymask
+                    return np.where(ymask if x is None else mask)
+                if mode == "subtract":
+                    mask = xmask & ~ymask
+                    return np.where(ymask if x is None else mask)
+            for scatter in fig.scatters:
+                x, y = fig.project(scatter.x, scatter.y, scatter.z)
+                mask = inside(x, y)
+                scatter.selected = join(scatter.selected, np.where(mask), fig.selection_mode)
+    fig.on_selection(lasso)
+
+
+
+def _make_triangles_lines(shape, wrapx=False, wrapy=False):
+    """Transform rectangular regular grid into triangles
+
+    :param x: {x2d}
+    :param y: {y2d}
+    :param z: {z2d}
+    :param bool wrapx: when True, the x direction is assumed to wrap, and polygons are drawn between the end end begin points
+    :param bool wrapy: simular for the y coordinate
+    :return: triangles and lines used to plot Mesh
+    """
+
+    nx, ny = shape
+
+    mx = nx if wrapx else nx - 1
+    my = ny if wrapy else ny - 1
+
+    """
+    create all pair of indices (i,j) of the rectangular grid
+    minus last row if wrapx = False => mx
+    minus last column if wrapy = False => my
+    |  (0,0)   ...   (0,j)    ...   (0,my-1)  |
+    |    .      .      .       .       .      |
+    |  (i,0)   ...   (i,j)    ...   (i,my-1)  |
+    |    .      .      .       .       .      |
+    |(mx-1,0)  ...  (mx-1,j)  ... (mx-1,my-1) |
+    """
+    i, j = np.mgrid[0:mx, 0:my]
+
+    """
+    collapsed i and j in one dimensional array, row-major order
+    ex :
+    array([[0,  1,  2],     =>   array([0, 1, 2, 3, *4*, 5])
+           [3, *4*, 5]])
+    if we want vertex 4 at (i=1,j=1) we must transform it in i*ny+j = 4
+    """
+    i, j = np.ravel(i), np.ravel(j)
+
+    """
+    Let's go for the triangles :
+        (i,j)    -  (i,j+1)   -> y dir
+        (i+1,j)  - (i+1,j+1)
+          |
+          v
+        x dir
+
+    in flatten coordinates:
+        i*ny+j     -  i*ny+j+1
+        (i+1)*ny+j -  (i+1)*ny+j+1
+    """
+
+    t1 = (i * ny + j,
+          (i + 1) % nx * ny + j,
+          (i + 1) % nx * ny + (j + 1) % ny)
+    t2 = (i * ny + j,
+          (i + 1) % nx * ny + (j + 1) % ny,
+          i * ny + (j + 1) % ny)
+
+    """
+        %nx and %ny are used for wrapx and wrapy :
+        if (i+1)=nx => (i+1)%nx=0 => close mesh in x direction
+        if (j+1)=ny => (j+1)%ny=0 => close mesh in y direction
+    """
+
+    nt = len(t1[0])
+
+    triangles = np.zeros((nt * 2, 3), dtype=np.uint32)
+    triangles[0::2, 0], triangles[0::2, 1], triangles[0::2, 2] = t1
+    triangles[1::2, 0], triangles[1::2, 1], triangles[1::2, 2] = t2
+
+    lines = np.zeros((nt * 4, 2), dtype=np.uint32)
+    lines[::4,0], lines[::4,1] = t1[:2]
+    lines[1::4,0], lines[1::4,1] = t1[0],t2[2]
+    lines[2::4,0], lines[2::4,1] = t2[2:0:-1]
+    lines[3::4,0], lines[3::4,1] = t1[1],t2[1]
+
+    return triangles, lines
